/**************************************************************************************
 *                                                                                    *
 * This application contains code from OpenDivX and is released as a "Larger Work"    *
 * under that license. Consistant with that license, this application is released     *
 * under the GNU General Public License.                                              *
 *                                                                                    *
 * The OpenDivX license can be found at: http://www.projectmayo.com/opendivx/docs.php *
 * The GPL can be found at: http://www.gnu.org/copyleft/gpl.html                      *
 *                                                                                    *
 * Copyright (c) 2001 - Project Mayo                                                  *
 *                                                                                    *
 * Authors: Damien Chavarria <adrc at projectmayo.com>                                *
 *                                                                                    *
 **************************************************************************************/


#include "AudioCodec.h"

/*
 * Class functions
 */

/*
 * Private wrapper 
 * around my C functions
 *
 */

int AudioCodec::DecompressMp3(char *outmemory, int outmemsize, int *done)
{

	if(this->last_result == MP3_OK) {
		this->last_result = decodeMP3(&this->mp, NULL, 0, outmemory, outmemsize, done);
    
    if(this->last_result == MP3_NEED_MORE) {
     
      if( this->decaps->ReadAudio(this->in_buffer, 16384) != -1) {
		
		  this->last_result = decodeMP3(&this->mp, this->in_buffer, 16384, outmemory, outmemsize, done);
		  return 1;
      }
      else {
	
		  return 0;
      }
    }
    else {
      return 1;
    }

  }
  else {
    
    if( this->decaps->ReadAudio(this->in_buffer, 16384) != -1) {
      this->last_result = decodeMP3(&this->mp, this->in_buffer, 16384, outmemory, outmemsize, done);
      return 1;
    }
    else {
      
      return 0;
    }
  }
}

/*
 * Constructor : tries to init the stream
 *
 */

AudioCodec::AudioCodec(AviDecaps *decaps, WAVEFORMATEX *lpWave)
{
	this->mpeg        = 0;
	this->in_buffer   = NULL;
	this->out_buffer  = NULL;

	if(lpWave) {

		char  msg[256];
		this->decaps     = decaps;

		if(lpWave->wFormatTag == 85) {
	
			InitMP3(&this->mp);
	  
		    this->last_result = MP3_NEED_MORE;
	        ring_init();

			this->in_buffer  = (char *) malloc(INPUT_BUFFER_SIZE);
			this->out_buffer = (char *) malloc(65536);

			if(this->DecompressMp3(this->out_buffer, 16384, &this->real_size) == MP3_ERR) {
			    return;
			}
	
			/*
			 * Buffering
			 */

			while(!ring_full(this->real_size))
				{
					this->DecompressMp3(this->out_buffer, 16384, &this->real_size);
			        ring_write(this->out_buffer, this->real_size);
				}

			this->mpeg    = 1;

			/*
			 * and sets up the output
			 * format for the renderer
			 */

			this->oFormat = (WAVEFORMATEX *) malloc(sizeof(WAVEFORMATEX));
			memcpy(this->oFormat, lpWave, sizeof(WAVEFORMATEX));
			
			this->oFormat->wFormatTag = WAVE_FORMAT_PCM;

			if (oFormat->wBitsPerSample != 8 && oFormat->wBitsPerSample != 16)
				oFormat->wBitsPerSample = 16;

			if (oFormat->nChannels!=1 && oFormat->nChannels!=2)
				oFormat->nChannels = 2;

			oFormat->nBlockAlign		= (oFormat->wBitsPerSample/8) * oFormat->nChannels;
			oFormat->nAvgBytesPerSec	=  oFormat->nBlockAlign * oFormat->nSamplesPerSec;
			oFormat->cbSize			    =  0;

		}
	}
}

/*
 * Destructor  : Cleanup.
 *
 */

AudioCodec::~AudioCodec()
{
	if(this->in_buffer != NULL) {

		free(this->in_buffer);
	}

	if(this->out_buffer) {

		free(this->out_buffer);
	}
}

/*
 * Return TRUE is codec is
 * ready to decompress
 */

BOOL AudioCodec::IsOK()
{
	return (this->mpeg == 1);
}

/*
 * Return the output format
 * (used by the renderer)
 *
 */

WAVEFORMATEX *AudioCodec::GetFormat()
{
	return this->oFormat;
}

/*
 * Empty all buffers
 *
 */

int AudioCodec::EmptyBuffers()
{
	ring_init();
	this->last_result = MP3_NEED_MORE;
	ExitMP3(&this->mp);
	InitMP3(&this->mp);

	return 1;
}


/*
 * decompress size octets of
 * audio to buffer
 *
 */

int AudioCodec::Decompress(void *buffer, int size)
{

	if(this->IsOK()) {

		char msg[256];
		int  i;

		if(size == 0)
			return 1;

		if(size < 4096) {

			/*
			 * we only read once
			 */

			while(!ring_full(this->real_size)) {
    
				if(this->DecompressMp3(this->out_buffer, 16384, &this->real_size) == 1) {
					ring_write(this->out_buffer, this->real_size);
				}
				else {
					return 0;
				}
			}
  
			ring_read((char *) buffer, size);
		}
		else {

			int blocks = size / 4096;

			for(i = 0; i < blocks; i++) {
			
				while(!ring_full(this->real_size)) {
    
					if(this->DecompressMp3(this->out_buffer, 16384, &this->real_size) == 1) {
						ring_write(this->out_buffer, this->real_size);
					}
					else {
						return 0;
					}
				}
  
				ring_read(((char *) buffer) + i*4096, 4096);
			}

			/*
			 * the last bit
			 */
			
			int left = size - (blocks * 4096);

			if(left > 0)
				ring_read(((char *) buffer) + (blocks - 1)*4096, left);

		}
	}

	return 1;
}

/*
 * Closes the decding engine
 *
 */

int AudioCodec::Close()
{
	if(this->mpeg) {

		ExitMP3(&this->mp);
	}

	return 1;
}