/**************************************************************************************
 *                                                                                    *
 * This application contains code from OpenDivX and is released as a "Larger Work"    *
 * under that license. Consistant with that license, this application is released     *
 * under the GNU General Public License.                                              *
 *                                                                                    *
 * The OpenDivX license can be found at: http://www.projectmayo.com/opendivx/docs.php *
 * The GPL can be found at: http://www.gnu.org/copyleft/gpl.html                      *
 *                                                                                    *
 * Copyright (c) 2001 - Project Mayo                                                  *
 *                                                                                    *
 * Authors: Damien Chavarria <adrc at projectmayo.com>                                *
 *                                                                                    *
 **************************************************************************************/


#include "InputMedia.h"

/*
 * HTTP filling thread
 */

DWORD WINAPI httpThreadFunc( LPVOID lpData)
{
	InputMedia *input = (InputMedia *) lpData;
	DWORD read;
	char  buffer[4096];

	while(input->running) {

		Sleep(5);

		if(input->buffering) {

			read = 0;
					
			InternetReadFile(input->http, (void *) buffer, 4096, &read);

			memcpy((void *) &input->httpBuffers[input->currentBuffer][input->downloaded], buffer, read);

			input->downloaded += read;

			if(input->downloaded >= HTTP_BUFFER_SIZE) {
			
				input->buffering = 0;
			}
		}
	}

	return 0;
}

/*
 * The main class
 */

InputMedia::InputMedia(char *lpFilename, int mode, int type) 
{
	if(lpFilename) {
	
		this->file       = NULL;
		this->hInternet  = NULL;		
		this->http       = NULL;
		this->buffering  = 0;
		this->mode       = -1;
		this->downloaded = 0;

		switch(type) {

		case INPUT_TYPE_FILE:

			switch(mode) {

			case INPUT_OPEN_BINARY:

				this->file = fopen(lpFilename, "rb");
				break;

			case INPUT_OPEN_ASCII:

				this->file = fopen(lpFilename, "rt");
				break;

			default:
			
				this->file = fopen(lpFilename, "rb");
				break;
			}

			this->mode = INPUT_TYPE_FILE;

			break;

		case INPUT_TYPE_HTTP:

			char *url;

			this->httpBuffers = (char **) malloc(2*sizeof(char *));

			this->httpBuffers[0] = (char *) malloc(HTTP_BUFFER_SIZE);
			this->httpBuffers[1] = (char *) malloc(HTTP_BUFFER_SIZE);

			this->hInternet = InternetOpen("The Playa - OpenDivX Viewer",
										   INTERNET_OPEN_TYPE_DIRECT,
										   NULL, NULL, 0);

			if(this->hInternet == NULL) {
				
				MessageBox(NULL, "Couldn't open an Internet connection", "", MB_OK);
				return;
			}

			url = (char *) malloc(strlen(lpFilename) + 7);
			sprintf(url, "http://%s", lpFilename);

			this->http = InternetOpenUrl(this->hInternet, url, NULL, -1, 0, 0);
			
			if(this->http == NULL) {

				PM_DEBUG("Couldn't open URL: %s", url);
				return;
			}

			/*
			 * Here we can create the temp file 
			 * and start the filling thread
			 */

			this->file = fopen("c:\\temp.avi", "w+b"); //tmpfile();

			if(this->file == NULL) {

				MessageBox(NULL, "Could not create temp file, check your settings.", "Error", MB_OK);
				this->file = NULL;
				return;
			}

			this->downloaded  = 0;
			this->lastReadPos = 0;

			this->ioMutex = CreateMutex (NULL, FALSE, NULL);

			this->buffering     = 1;
			this->running       = 1;
			this->currentBuffer = 0;
			this->httpThread    = CreateThread(NULL, 0, httpThreadFunc, (LPVOID) this, 0, &this->httpId);

			this->mode = INPUT_TYPE_HTTP;

			break;

			default:
				break;
		}
	}
}

InputMedia::~InputMedia() 
{
	if(this->file) {

		fclose(this->file);
	}
}

int InputMedia::isOK() {

	return (this->file != NULL);
}

int InputMedia::Read(char *data, unsigned int size)
{
	switch(this->mode) {

	case INPUT_TYPE_FILE:

		if(this->file) {

			return fread( (void *) data, 1, size, this->file);
		}

		MessageBox(NULL, "Problem reading from source!", "", MB_OK);
		break;

	case INPUT_TYPE_HTTP:
		
		if(this->http) {

			memcpy(data, (void *) &this->httpBuffers[this->currentBuffer][this->lastReadPos], size);
			
			this->lastReadPos += size;
			
			return size;
		}
		break;

	default:
		break;
	}

	return 0;
}

int InputMedia::ScanF(const char *format, void *data) {

	switch(mode) {

	case INPUT_TYPE_FILE:
	case INPUT_TYPE_HTTP:

		if(this->file) {
	
			fscanf(this->file, format, data);

			return 1;
		}
		break;

	default:
		break;
	}

	return 0;
}


int InputMedia::Seek(int size, unsigned int method)
{

	switch(mode) {

	case INPUT_TYPE_FILE:

		if(this->file) {

			switch(method)
			{
			case INPUT_SEEK_SET:

				return fseek( this->file, size, SEEK_SET);
				break;

			
			case INPUT_SEEK_CUR:

				if(size == 0) {
					return ftell(this->file);	
				}
				else {
					return fseek( this->file, size, SEEK_CUR);
				}
				break;

			case INPUT_SEEK_END:

				return fseek( this->file, size, SEEK_END);
				break;
			}
		}
		break;

	case INPUT_TYPE_HTTP:

		if(this->file) {

			switch(method)
			{
			case INPUT_SEEK_SET:

				this->lastReadPos = size;
				return 0; 
				break;

			
			case INPUT_SEEK_CUR:

				if(size == 0) {
					return this->lastReadPos;	
				}
				else {
					this->lastReadPos += size;
					return 0;
				}
				break;

			case INPUT_SEEK_END:

				break;
			}
		}
		break;

	default:
		break;
	}

	return 0;
}

int InputMedia::Close()
{
	if(this->file) {

		fclose(this->file);
	}

	if(this->http) {

		InternetCloseHandle(this->http);
	}

	if(this->hInternet) {

		InternetCloseHandle(this->hInternet);
	}

	return 1;
}