/**************************************************************************************
 *                                                                                    *
 * This application contains code from OpenDivX and is released as a "Larger Work"    *
 * under that license. Consistant with that license, this application is released     *
 * under the GNU General Public License.                                              *
 *                                                                                    *
 * The OpenDivX license can be found at: http://www.projectmayo.com/opendivx/docs.php *
 * The GPL can be found at: http://www.gnu.org/copyleft/gpl.html                      *
 *                                                                                    *
 * Copyright (c) 2001 - Project Mayo                                                  *
 *                                                                                    *
 * Authors: Damien Chavarria <adrc at projectmayo.com>                                *
 *                                                                                    *
 **************************************************************************************/


#include "Skin.h"
#include "Playback.h"
#include "Playlist.h"

#ifdef WIN32
#include "../build/win32/ressources.h"
#include "../build/win32/resource.h"
#endif

#include <math.h>
#include <commctrl.h>
#include <windows.h>
#include <windowsx.h>

/*
 * Intervals for the timers
 */

#define TIMER_ID     1
#define TIMER_RATE  50

/*
 * Variables
 *
 */

int       action  = ACTION_NONE;

DWORD     maintain_aspect_ratio = 1;
DWORD     use_bilinear          = 1;
DWORD     force_fullscreen_res  = 0;
DWORD     use_subtitles         = 1;
DWORD     no_resize             = 0;

UINT      uTimerID;
RECT      clientRect;
RECT      windowRect;
int       moveX = 0, moveY = 0;
Skin     *skin;
Playback *playback;
Playlist *playlist;
char      Name[] = "The \"Playa\"";
char     *RecentFiles[5];
HWND      hwnd, about;
HMENU     popupMenu;
POINT     pt2;
HINSTANCE hInstance;

/*
 * needed definitions
 */

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
void OpenFileForPlaying(HWND hwnd, char *filename, int type);

/*
 * Save and load options
 */

void LoadOptions() {

	HKEY     key;
	DWORD created, size;
	LONG     result;
	DWORD  type, i;
	char      file[256], name[5];

	/*
	 * Init the recent file list 
	 */

	for(i=0; i < 5; i++) {

		RecentFiles[i] = NULL;
	}

	/*
	 * Open the registry key
	 */

	result = RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\DivXNetworks\\ThePlaya",
							              0, "CONFIG", REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, 
							              NULL, &key, &created);

	if(result != ERROR_SUCCESS) {

		MessageBox(NULL, "Couldn't load options", "", MB_OK);
		return;
	}

	switch(created) {

	case REG_CREATED_NEW_KEY:

		/*
		 * First time launch (we keep the default)
		 * 
		 */

		break;

	case REG_OPENED_EXISTING_KEY:

		/*
		 * We can read the values
 		 */

		size = 4;
		result = RegQueryValueEx(key, "UseBilinear", 0, &type, (BYTE *)&use_bilinear, &size);
		size = 4;
		result = RegQueryValueEx(key, "MaintainAspectRatio", 0, &type, (BYTE *)&maintain_aspect_ratio, &size);
		size = 4;
		result = RegQueryValueEx(key, "ForceFullscreenRes", 0, &type, (BYTE *)&force_fullscreen_res, &size);
		size = 4;
		result = RegQueryValueEx(key, "UseSubtitles", 0, &type, (BYTE *)&use_subtitles, &size);

		/*
		 * And the recent files
		 */

		for(i=0; i < 5; i++) {

			sprintf(name, "File%d", i+1);

			size = 256;
			result = RegQueryValueEx(key, name, 0, &type, (BYTE *)&file, &size);

			if(result == ERROR_SUCCESS) {
			
				RecentFiles[i] = (char *) malloc(size);
				strncpy(RecentFiles[i], file, size);
			}
		}

		break;

	default:
		break;
	}

}

void SaveOptions() {

	HKEY     key;
	LONG     result;
	DWORD created, i;
	char      name[5];

	/*
	 * Try to open the registry key
	 */

	result = RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\DivXNetworks\\ThePlaya",
							              0, "CONFIG", REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, 
							              NULL, &key, &created);

	if(result != ERROR_SUCCESS) {

		MessageBox(NULL, "Couldn't save options", "", MB_OK);
		return;
	}

	RegSetValueEx(key, "UseBilinear", 0, REG_DWORD, (BYTE *) &use_bilinear, 4);
	RegSetValueEx(key, "MaintainAspectRatio", 0, REG_DWORD, (BYTE *) &maintain_aspect_ratio, 4);
	RegSetValueEx(key, "ForceFullscreenRes", 0, REG_DWORD, (BYTE *) &force_fullscreen_res, 4);
	RegSetValueEx(key, "UseSubtitles", 0, REG_DWORD, (BYTE *) &use_subtitles, 4);

	/*
	 * The recent file list
	 */

	for(i=0; i < 5; i++) {
		
		if(RecentFiles[i] != NULL) {
		
			sprintf(name, "File%d\0", i+1);
			RegSetValueEx(key, name, 0, REG_SZ, 
							      (BYTE *) RecentFiles[i], strlen(RecentFiles[i]));
		}
	}
}


/*
 * About dialog box message function
 *
 */

int APIENTRY AboutDlgProc (HWND hDlg, WORD wMsg, LONG wParam, LONG lParam)
{
  int first = 0;

  switch (wMsg)
    {
      case WM_INITDIALOG:
        return (0);
		break;

      case WM_SYSCOMMAND:
        if (wParam == SC_CLOSE)
          {
          EndDialog (hDlg, TRUE);
          return (TRUE);
          }
        break;

	  case WM_COMMAND:

        switch (wParam)
		{
		  case IDB_ABOUT_OK:
			
			  EndDialog(hDlg, TRUE);
			  return TRUE;
			  break;

		  default:
			  return 0;
		}
		break;
	}

  return FALSE;
}

/*
 * Open URL dialog box message function
 *
 */

int APIENTRY UrlDlgProc (HWND hDlg, WORD wMsg, LONG wParam, LONG lParam)
{

  switch (wMsg)
    {
      case WM_INITDIALOG:

  		  SendMessage(GetDlgItem(hDlg, IDC_TYPE_COMBO), CB_ADDSTRING, 0, (LONG)"http://");
	      SendMessage(GetDlgItem(hDlg, IDC_TYPE_COMBO), CB_SETCURSEL, 0, 0);
		  
		  return (0);
		break;

      case WM_SYSCOMMAND:

        if (wParam == SC_CLOSE) {

            EndDialog (hDlg, TRUE);
            return (TRUE);
        }
        break;

	  case WM_COMMAND:

        switch (wParam)
		{
		  case IDB_URL_OK:
			
			  //OpenFileForPlaying(hwnd, "", INPUT_TYPE_HTTP);

			  EndDialog(hDlg, TRUE);
			  return TRUE;
			  break;

		  case IDB_URL_CANCEL:
			
			  EndDialog(hDlg, TRUE);
			  return TRUE;
			  break;

		  default:
			  return 0;
		}
		break;
	}

  return FALSE;
}


/*
 * Config dialog box message function
 *
 */

int APIENTRY OptionsDlgProc (HWND hDlg, WORD wMsg, LONG wParam, LONG lParam)
{
  int first = 0;

  switch (wMsg)
    {
      case WM_INITDIALOG:

		  CheckDlgButton(hDlg, IDC_CHECK_BILINEAR, use_bilinear);
		  CheckDlgButton(hDlg, IDC_CHECK_MAINTAIN, maintain_aspect_ratio);
		  CheckDlgButton(hDlg, IDC_CHECK_FORCE, force_fullscreen_res);
		  CheckDlgButton(hDlg, IDC_SUBTITLES, use_subtitles);
		  
		  SendMessage(GetDlgItem(hDlg, IDC_COMBO_RES), CB_ADDSTRING, 0, (LONG)"640x480");
		  SendMessage(GetDlgItem(hDlg, IDC_COMBO_RES), CB_ADDSTRING, 0, (LONG)"800x600");
		  SendMessage(GetDlgItem(hDlg, IDC_COMBO_RES), CB_ADDSTRING, 0, (LONG)"1024x768");
		  SendMessage(GetDlgItem(hDlg, IDC_COMBO_RES), CB_ADDSTRING, 0, (LONG)"1280x1024");
		  SendMessage(GetDlgItem(hDlg, IDC_COMBO_RES), CB_ADDSTRING, 0, (LONG)"1600x1200");

		  /*
		   * Check if we have a selection or not
		   */

		  if(force_fullscreen_res > 0) {
			
		      SendMessage(GetDlgItem(hDlg, IDC_COMBO_RES), CB_SETCURSEL, force_fullscreen_res - 1, 0);
		  }
		  else {

			  EnableWindow(GetDlgItem(hDlg, IDC_COMBO_RES), FALSE);
		  }

		  return (0);
		  break;

      case WM_SYSCOMMAND:
        if (wParam == SC_CLOSE)
          {
          EndDialog (hDlg, TRUE);
          return (TRUE);
          }
        break;

	  case WM_COMMAND:

        switch (wParam)
		{

		case IDB_OPTIONS_OK:

			/*
			 * Save the options here
			 */

			use_bilinear = IsDlgButtonChecked(hDlg, IDC_CHECK_BILINEAR);
			maintain_aspect_ratio = IsDlgButtonChecked(hDlg, IDC_CHECK_MAINTAIN);
			use_subtitles = IsDlgButtonChecked(hDlg, IDC_SUBTITLES);
		    
			CheckMenuItem(popupMenu, (UINT)ID_BILINEAR,     use_bilinear ? MF_CHECKED : MF_UNCHECKED);
			CheckMenuItem(popupMenu, (UINT)ID_ASPECT_RATIO, maintain_aspect_ratio ? MF_CHECKED : MF_UNCHECKED);
			
			/*
			 * Get the desired fullscreen resolution (if specified)
			 */

			if(force_fullscreen_res > 0) {
				
				force_fullscreen_res = (DWORD) SendMessage(GetDlgItem(hDlg, IDC_COMBO_RES),
																					 CB_GETCURSEL, 0, 0L);
				force_fullscreen_res++;

				if(force_fullscreen_res == CB_ERR) {
				
					force_fullscreen_res = 0;
				}
			}

			SaveOptions();

			EndDialog(hDlg, TRUE);
			break;

		case IDB_OPTIONS_CANCEL:
		    
			EndDialog(hDlg, TRUE);
			break;

		case IDC_CHECK_FORCE:

			if(force_fullscreen_res > 0) {

				force_fullscreen_res = 0;
				EnableWindow(GetDlgItem(hDlg, IDC_COMBO_RES), FALSE);
			}
			else {

				force_fullscreen_res = 255;
				EnableWindow(GetDlgItem(hDlg, IDC_COMBO_RES), TRUE);
			}
			break;
	
		default:
			  return 0;
		}
		break;
	}

  return FALSE;
}

/*
 * Main function
 */

int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInstance, LPSTR lpszCmdParam, int nCmdShow)
{ 
	POINT pt;
    MSG Msg;
    WNDCLASS W;
	DWORD i;

	hInstance = hInst;

	/*
	 * Load the options
	 *
	 */

	LoadOptions();

	/*
     * init the video subsystem 
     *
     */
	
   playback = new Playback();
   playlist = new Playlist();

   /*
    * Loads the skin
    *
    */

   skin = new Skin(hInst);

   clientRect.left   = 0;
   clientRect.right  = skin->DefaultWidth();
   clientRect.top    = 0;
   clientRect.bottom = skin->DefaultHeight();
   
   windowRect.left   = 0;
   windowRect.right  = skin->DefaultWidth();
   windowRect.top    = 0;
   windowRect.bottom = skin->DefaultHeight();

   AdjustWindowRect(&windowRect, WS_OVERLAPPEDWINDOW, 0);

   /*
    * Window Class
    */

   memset(&W, 0, sizeof(WNDCLASS));
   
   W.style         = CS_HREDRAW | CS_VREDRAW;
   W.lpfnWndProc   = WndProc;
   W.hInstance     = hInst;
   W.hbrBackground = (HBRUSH)(0);
   W.hCursor       = LoadCursor(hInst, MAKEINTRESOURCE(IDC_CURSOR1));
   W.hIcon         = LoadIcon(hInst, MAKEINTRESOURCE(IDB_ICON));
   W.lpszClassName = Name;
   W.lpszMenuName  = NULL;


   RegisterClass(&W);

   /*
    * Load the menu and change 
	* it for recent file list
    */

   HMENU             menu;
   MENUITEMINFO itemInfo;

   popupMenu = LoadMenu(hInst, MAKEINTRESOURCE(IDR_MENU1));

   menu = GetSubMenu(popupMenu, 0);
   menu = GetSubMenu(menu, 5);
   
   if(RecentFiles[0] != NULL) {

	   RemoveMenu(menu, 0, MF_BYPOSITION);
   
		/*
		 * Now add the files
		 */
   
	   for(i=0; i < 5; i++) {
	   
		   if(RecentFiles[i] != NULL) {

			   itemInfo.cbSize = sizeof(MENUITEMINFO);
			   itemInfo.fMask  = MIIM_DATA | MIIM_TYPE | MIIM_ID;
			   itemInfo.fType  = MFT_STRING;
			   itemInfo.dwTypeData = RecentFiles[i];
			   itemInfo.cch = strlen(RecentFiles[i]);
			   itemInfo.wID = ID_RECENT_FILE1 + i;
		   
			   InsertMenuItem(menu, GetMenuItemCount(menu), TRUE, &itemInfo);
			}
	   }
   }


   /*
    * Create the main window
    *
    */

   hwnd = CreateWindow( Name, Name, WS_OVERLAPPEDWINDOW, 100, 30, 
						windowRect.right - windowRect.left, 
						windowRect.bottom - windowRect.top, 
						NULL, NULL, hInst, NULL);
   
   /*
    * Set the window Rgn
    *
    */

		GetClientRect(hwnd, &clientRect);
		GetWindowRect(hwnd, &windowRect);

		pt.x = clientRect.left;
		pt.y = clientRect.top;

		ClientToScreen(hwnd, &pt);

		pt2.x = clientRect.right;
		pt2.y = clientRect.bottom;

		ClientToScreen(hwnd, &pt2);

		SetWindowRgn(hwnd, CreateRectRgn( pt.x  - windowRect.left, 
			  							  pt.y  - windowRect.top,
										  (windowRect.right - windowRect.left) - (windowRect.right - pt2.x),
										  (windowRect.bottom - windowRect.top) - (windowRect.bottom - pt2.y)), TRUE); 
   /*
    * now The buttons
    */

   ShowWindow(hwnd,nCmdShow);
   UpdateWindow(hwnd);

   /*
    * Register for Drag n' Drop
    */

   DragAcceptFiles(hwnd, TRUE);

   /*
    * And Start TIMER
    *
    */

	SetTimer(hwnd, TIMER_ID, TIMER_RATE, NULL);

	/*
	 * Set the icon
	 */

	SetClassLong(hwnd, GCL_HICON, (LONG) LoadIcon(hInst, MAKEINTRESOURCE(IDB_ICON))); 
 
	/*
	 * Menu items 
	 */

   CheckMenuItem(popupMenu, (UINT)ID_BILINEAR,     use_bilinear ? MF_CHECKED : MF_UNCHECKED);
   CheckMenuItem(popupMenu, (UINT)ID_ASPECT_RATIO, maintain_aspect_ratio ? MF_CHECKED : MF_UNCHECKED);

   while (TRUE)
    {
        if (PeekMessage(&Msg, NULL, 0, 0, PM_NOREMOVE))
        {
            if (!GetMessage(&Msg, NULL, 0, 0))
                return (int) Msg.wParam;

            TranslateMessage(&Msg);
            DispatchMessage(&Msg);
		}
		else {

		}
   }

   return Msg.wParam;
}

/*
 * Opens a specified file for playing
 */

void OpenFileForPlaying(HWND hwnd, char *filename, int type) {

	playback->Close();

	skin->Display(hwnd, playback);

	RECT rect, windowrect;
					
	playback->Close();
	if(playback->Open(filename, type, hwnd)) {

	/*
	 * First resize the window
	 */
					
	if(playback->OK()) {

		DWORD i;

		GetWindowRect(hwnd, &windowrect); 
		rect.left   = windowrect.left;
		rect.top    = windowrect.top;
		rect.right  = playback->Width() + 15;
		rect.bottom = playback->Height() + 115 + 22 + (playback->has_subtitles && use_subtitles ? 50 : 0);

		MoveWindow( hwnd, rect.left, rect.top, rect.right, rect.bottom, TRUE);

		/*
		 * Update flags
		 */

		playback->SetBilinear(use_bilinear);
		playback->SetForceFullscreen(force_fullscreen_res);
		playback->SetUseSubtitles(use_subtitles);

		/*
		 * Update recent file list
		 */

		DWORD insert = 1;

		for(i=0; i < 5; i++) {
				
		if(RecentFiles[i] != NULL && strcmp(RecentFiles[i], filename) == 0) {
				insert = 0;
			}
		}

		if(insert) {
		
			for(i=4; i > 0; i--) {
				
				RecentFiles[i] = RecentFiles[i - 1];
			}

			RecentFiles[0] = (char *) malloc(strlen(filename) + 1);
			strncpy(RecentFiles[0], filename, strlen(filename));
			RecentFiles[0][strlen(filename)] = '\0';
		}

		/*
		 * And update the menu
		 */

		HMENU menu, sub;

		menu = GetSubMenu(popupMenu, 0);
		
		/*
		 * The subtitles
		 */

		if(playback->has_subtitles) {

			sub = GetSubMenu(menu, 16);
			RemoveMenu(sub, 0, MF_BYPOSITION);				
		
			MENUITEMINFO subitemInfo;

		    subitemInfo.cbSize = sizeof(MENUITEMINFO);
			subitemInfo.fMask  = MIIM_DATA | MIIM_TYPE | MIIM_ID;
			subitemInfo.fType  = MFT_STRING;
			subitemInfo.dwTypeData = "Use Subtitles";
			subitemInfo.cch = 13;
			subitemInfo.wID = ID_SUBTITLES;
		   
		    InsertMenuItem(sub, 0, TRUE, &subitemInfo);
		    CheckMenuItem(sub, (UINT)ID_SUBTITLES, use_subtitles ? MF_CHECKED : MF_UNCHECKED);
		}
		else {

			sub = GetSubMenu(menu, 16);
			RemoveMenu(sub, 0, MF_BYPOSITION);				
		
			MENUITEMINFO subitemInfo;

		    subitemInfo.cbSize = sizeof(MENUITEMINFO);
			subitemInfo.fMask  = MIIM_DATA | MIIM_TYPE | MIIM_ID;
			subitemInfo.fType  = MFT_STRING;
			subitemInfo.dwTypeData = "None Available";
			subitemInfo.cch = 14;
			subitemInfo.wID = ID_NONE_SUB;
		   
		    InsertMenuItem(sub, 0, TRUE, &subitemInfo);
			EnableMenuItem(sub, (UINT)ID_NONE_SUB, MF_GRAYED);
		}

		/*
		 * And the recent file list
		 */
		
		menu = GetSubMenu(menu, 5);

		if(RecentFiles[0] != NULL) {

			DWORD count;
			MENUITEMINFO itemInfo;

			count = GetMenuItemCount(menu);

			for(i=0; i < count; i++) {
						
				RemoveMenu(menu, 0, MF_BYPOSITION);				
			}

			/*
			 * Now add the files
			 */
   
		    for(i=0; i < 5; i++) {
	   
			   if(RecentFiles[i] != NULL) {

				  itemInfo.cbSize = sizeof(MENUITEMINFO);
				  itemInfo.fMask  = MIIM_DATA | MIIM_TYPE | MIIM_ID;
				  itemInfo.fType  = MFT_STRING;
				  itemInfo.dwTypeData = RecentFiles[i];
				  itemInfo.cch = strlen(RecentFiles[i]);
				  itemInfo.wID = ID_RECENT_FILE1 + i;
		   
				  InsertMenuItem(menu, GetMenuItemCount(menu), TRUE, &itemInfo);
			   }
		   }
		}

		/*
	 	 * Then play
		 */
		
		playback->Play();
	}
	}
}

/*
 * Opens a file:
 */

void fileOpen(HWND hwnd)
{
	OPENFILENAME ofn;
	char szFile[260];
				
	/*
	 * shows a file selector
	 */

	ZeroMemory(&ofn, sizeof(OPENFILENAME));
    ZeroMemory(szFile, 260);

	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hwnd;
	ofn.lpstrFile = szFile;
	ofn.nMaxFile = sizeof(szFile);

	ofn.lpstrFilter = "All\0*.*\0AVI Files\0*.AVI\0\0";
	ofn.nFilterIndex = 2;
	ofn.lpstrFileTitle = "";
	ofn.nMaxFileTitle = 0;
	ofn.lpstrInitialDir = NULL;
	ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST/* | OFN_ALLOWMULTISELECT | OFN_EXPLORER*/;
				
	if (GetOpenFileName(&ofn) == TRUE) {
		
		OpenFileForPlaying(hwnd, ofn.lpstrFile, INPUT_TYPE_FILE);
	}
}

/*
 * Handles events
 */

LRESULT CALLBACK WndProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{ 
	int xPos, yPos;

	switch(Message) {

	case WM_DESTROY:

		SaveOptions();
		playback->Close();
		PostQuitMessage(0);
		return 0;
		break;

	case WM_SYSCOMMAND:
        
		if (wParam == SC_CLOSE)
          {
			return (TRUE);
          }
        break;

	case WM_COMMAND	:
    
		switch(LOWORD(wParam)) {

		case ID_SUBTITLES:

			RECT  wrect;
			HMENU menu;

			menu = GetSubMenu(popupMenu, 0);
			menu = GetSubMenu(menu, 16);

			if(playback != NULL) {

				GetWindowRect(hwnd, &wrect);

				if(use_subtitles) {

					use_subtitles = 0;
					wrect.bottom -= 50;

					MoveWindow(hwnd, wrect.left, wrect.top, wrect.right - wrect.left, wrect.bottom - wrect.top, TRUE);
					CheckMenuItem(menu, (UINT)ID_SUBTITLES, MF_UNCHECKED);

					playback->SetUseSubtitles(use_subtitles);
				}
				else {

					use_subtitles = 1;
					wrect.bottom += 50;

					CheckMenuItem(menu, (UINT)ID_SUBTITLES, MF_CHECKED);
					MoveWindow(hwnd, wrect.left, wrect.top, wrect.right - wrect.left, wrect.bottom - wrect.top, TRUE);

					playback->SetUseSubtitles(use_subtitles);
				}
			}

			break;

		case ID_RECENT_FILE1:

			OpenFileForPlaying(hwnd, RecentFiles[0], INPUT_TYPE_FILE);
			break;

		case ID_RECENT_FILE2:

			OpenFileForPlaying(hwnd, RecentFiles[1], INPUT_TYPE_FILE);
			break;

		case ID_RECENT_FILE3:

			OpenFileForPlaying(hwnd, RecentFiles[2], INPUT_TYPE_FILE);
			break;

		case ID_RECENT_FILE4:

			OpenFileForPlaying(hwnd, RecentFiles[3], INPUT_TYPE_FILE);
			break;

		case ID_RECENT_FILE5:

			OpenFileForPlaying(hwnd, RecentFiles[4], INPUT_TYPE_FILE);
			break;

		case ID_MENU_ABOUT:

            DialogBox (hInstance, (LPCSTR)MAKEINTRESOURCE(IDD_ABOUT_DIALOG),
                        hwnd, (DLGPROC)AboutDlgProc);
						
			break;

		case ID_MENU_OPTIONS:

            DialogBox (hInstance, (LPCSTR)MAKEINTRESOURCE(IDD_OPTION_DIALOG),
                             hwnd, (DLGPROC)OptionsDlgProc);
						
			break;

		case ID_OPEN_URL:

            DialogBox (hInstance, (LPCSTR)MAKEINTRESOURCE(IDD_URL_DIALOG),
                             hwnd, (DLGPROC)UrlDlgProc);
			break;

		case ID_OPEN_FILE:
			fileOpen(hwnd);
			break;

		case ID_PREVIOUS:

			playback->Rewind();
			break;

		case ID_STOP:

			RECT rect;

			playback->Stop(TRUE);

			GetClientRect(hwnd, &rect);
			InvalidateRect(hwnd, &rect, TRUE); 
 			UpdateWindow(hwnd);
			break;
		
		case ID_PLAY:

			if(playback->OK()) {
				
				if(playback->isPaused()) {
					playback->Pause();
				}
				else {
					playback->Play();
				}
			}
			else {

				/*
				 * Load files
				 */
			
				fileOpen(hwnd);
			}

			break;
		
		case ID_PAUSE:

			if(playback->OK()) {

				if(playback->isPaused()) {
				
					playback->NextFrame();
				}
				else {
					playback->Pause();
				}
			}

			break;

		case ID_NEXT:

			playback->FastForward();
			break;

			
		case ID_EXIT:

			SaveOptions();
		    DestroyWindow(hwnd);
			break;

		case ID_FULLSCREEN:

			GetWindowRect(hwnd, &windowRect);

			playback->SetForceFullscreen(force_fullscreen_res);
			playback->Fullscreen(TRUE);
			SetCursor(NULL);
			break;

		case ID_ORIGINAL_SIZE:

			if(playback->isPlaying()) {

				RECT src;

				GetWindowRect(hwnd, &windowRect);

				src.left    = 0;
				src.right   = playback->Width() + 15;
				src.top     = 0;
				src.bottom  = playback->Height() + 22 + 115;

				AdjustWindowRect(&src, WS_OVERLAPPEDWINDOW, 0);

				MoveWindow( hwnd, windowRect.left, 
							windowRect.top, 
							src.right - src.left, 
							src.bottom - src.top, TRUE);
			}
			break;

			case ID_HALF_SIZE:

			if(playback->isPlaying()) {

				RECT src;

				GetWindowRect(hwnd, &windowRect);

				src.left    = 0;
				src.right   = playback->Width()/2 + 15;
				src.top     = 0;
				src.bottom  = playback->Height()/2 + 22 + 115;

				AdjustWindowRect(&src, WS_OVERLAPPEDWINDOW, 0);

				MoveWindow( hwnd, windowRect.left, 
							windowRect.top, 
							src.right - src.left, 
							src.bottom - src.top, TRUE);
			}
			break;

			case ID_DOUBLE_SIZE:

			if(playback->isPlaying()) {

				RECT src;

				GetWindowRect(hwnd, &windowRect);

				src.left    = 0;
				src.right   = playback->Width()*2 + 15;
				src.top     = 0;
				src.bottom  = playback->Height()*2 + 22 + 115;

				AdjustWindowRect(&src, WS_OVERLAPPEDWINDOW, 0);

				MoveWindow( hwnd, windowRect.left, 
							windowRect.top, 
							src.right - src.left, 
							src.bottom - src.top, TRUE);
			}
			break;

			case ID_BILINEAR:

				use_bilinear = use_bilinear == 1 ? 0 : 1;				
				CheckMenuItem(popupMenu, (UINT)wParam, use_bilinear ? MF_CHECKED : MF_UNCHECKED);

				playback->SetBilinear(use_bilinear);

				break;

			case ID_ASPECT_RATIO:

				maintain_aspect_ratio = maintain_aspect_ratio == 1 ? 0 : 1;				
				CheckMenuItem(popupMenu, (UINT)wParam, maintain_aspect_ratio ? MF_CHECKED : MF_UNCHECKED);
				break;
		}
	break;

	/*
	 * Timer
	 *
	 */

	case WM_TIMER:

		if(playback && playback->isPlaying() && !playback->fullscreen) {

			if(action != ACTION_PROGRESS_CURSOR) {
				skin->SetProgressValue(hwnd, playback->GetProgress());
			}

			//skin->DisplayFps(hwnd, playback);

			/*
			 * Do time display here
			 */
		}

		break;

	/*
	 * Key Press
	 *
	 */

	case WM_KEYUP:

		switch(wParam) {
			
			case VK_ESCAPE:

				if(playback->fullscreen) {

					playback->Fullscreen(FALSE);
					SetCursor(LoadCursor(hInstance, MAKEINTRESOURCE(IDC_CURSOR1)));

					/*
					 * We need to resize the window to it's original size
					 */

					MoveWindow(hwnd, windowRect.left, windowRect.top, windowRect.right - windowRect.left, windowRect.bottom - windowRect.top, TRUE);
				}
				break;

			case 13:

				GetWindowRect(hwnd, &windowRect);

				playback->SetForceFullscreen(force_fullscreen_res);
				playback->Fullscreen(TRUE);
				SetCursor(NULL);
				break;

			case VK_SPACE:

				use_bilinear = use_bilinear == 1 ? 0 : 1;				
				CheckMenuItem(popupMenu, (UINT)wParam, use_bilinear ? MF_CHECKED : MF_UNCHECKED);

				playback->SetBilinear(use_bilinear);
				break;


		default:
			break;
		}
		break;

	case WM_LBUTTONDOWN:

		xPos = LOWORD(lParam); 
		yPos = HIWORD(lParam);

		action = skin->GetAction(xPos, yPos);
		skin->SetActionStart(hwnd, action);

		if(action == ACTION_RESIZE) {

			POINT pt;

			pt.x = xPos;
			pt.y = yPos;

			ClientToScreen(hwnd, &pt);

			moveX = pt.x;
			moveY = pt.y;

			GetWindowRect(hwnd, &windowRect);

			SetCapture(hwnd);

			if(!playback->isPaused())
				playback->Pause();	
		}

		if(action == ACTION_NONE) {

			/*
			 * We are moving the window
			 */

			action = ACTION_MOVING;

			moveX = xPos;
			moveY = yPos;

			GetWindowRect(hwnd, &windowRect);

			SetCapture(hwnd);
			
			if(!playback->isPaused())
				playback->Pause();
		}

		if(action == ACTION_VOLUME_CURSOR || action == ACTION_PROGRESS_CURSOR) {

			moveX = xPos;
		}

		break;

	case WM_LBUTTONUP:

		skin->SetActionEnd(hwnd, action);

		switch(action) {

		case ACTION_MENU:

			HMENU menu;

			POINT pt;

			pt.x = pt.y = 0;
			ClientToScreen(hwnd, &pt);

			xPos = pt.x + LOWORD(lParam); 
			yPos = pt.y + HIWORD(lParam);

			menu = GetSubMenu(popupMenu, 0);

			TrackPopupMenu( menu, 0, xPos, yPos, 0, hwnd, NULL);
			break;

		case ACTION_CLOSE:
			SaveOptions();
			
			if(playback != NULL)
				playback->Close();
			
			PostQuitMessage(0);
			break;

		case ACTION_MINIMIZE:

			ShowWindow(hwnd, SW_MINIMIZE);
			break;

		case ACTION_STOP:

			playback->Stop(TRUE);
			break;

		case ACTION_PLAY:

			if(playback->OK()) {
				
				if(playback->isPaused()) {
					playback->Pause();
				}
				else {
					playback->Play();
				}
			}
			else {

				/*
				 * Load files
				 */
			
				fileOpen(hwnd);
			}
			break;

		case ACTION_PAUSE:

			if(playback->OK()) {

				if(playback->isPaused()) {
				
					playback->NextFrame();
				}
				else {
					playback->Pause();
				}
			}
			break;

		case ACTION_FORWARD:

			playback->FastForward();
			break;

		case ACTION_REWIND:

			playback->Rewind();
			break;

		case ACTION_LOAD:
			{
				fileOpen(hwnd);
			}
			break;

		case ACTION_PROGRESS_CURSOR:

			playback->Seek(skin->GetProgress());
			break;

		case ACTION_VOLUME_CURSOR:

			playback->SetVolume(skin->GetVolume());
			break;

		case ACTION_VOLUME_BAR:

			moveX = LOWORD(lParam); 
			skin->SetVolumeCursorX(hwnd, moveX);

			playback->SetVolume(skin->GetVolume());
		
			break;

		case ACTION_PROGRESS_BAR:

			moveX = LOWORD(lParam); 
			skin->SetProgressCursorX(hwnd, moveX);
			playback->Seek(skin->GetProgress());
			break;

		case ACTION_NONE:
		default:
			break;
		}

		if(action == ACTION_MOVING || action == ACTION_RESIZE) {
			ReleaseCapture();
			RECT rect;

			GetWindowRect(hwnd, &rect);

			SetWindowPos(hwnd, HWND_TOP, rect.left, rect.top, 
						 rect.right - rect.left, rect.bottom - rect.top, 
						 SWP_SHOWWINDOW | SWP_NOZORDER | SWP_NOACTIVATE);

			if(playback->isPaused())
				playback->Pause();

			GetClientRect(hwnd, &rect);
			InvalidateRect(hwnd, &rect, TRUE); 
 			UpdateWindow(hwnd);
		}

		action = ACTION_NONE;

		break;

	case WM_MOUSEMOVE:

		if(action == ACTION_MOVING) {
			
			POINT pt;

			xPos = LOWORD(lParam); 
			yPos = HIWORD(lParam);

			pt.x = xPos;
			pt.y = yPos;

			ClientToScreen(hwnd, &pt);

			MoveWindow( hwnd, pt.x - moveX, 
						pt.y - moveY - 22, 
						windowRect.right - windowRect.left, 
						windowRect.bottom - windowRect.top, TRUE);
		}

		if(action == ACTION_RESIZE) {

			POINT pt;

			xPos = LOWORD(lParam); 
			yPos = HIWORD(lParam);

			pt.x = xPos;
			pt.y = yPos;

			ClientToScreen(hwnd, &pt);

			if(abs(pt.x - moveX) > 15 || abs(pt.y - moveY) > 15) {
			
			if(!playback->isPlaying()) {

				if(windowRect.right + (pt.x - moveX) - windowRect.left >= skin->DefaultWidth() &&
				   windowRect.bottom + (pt.y - moveY) - windowRect.top >= skin->DefaultHeight()) {
					MoveWindow( hwnd, windowRect.left, 
								windowRect.top, 
								windowRect.right + (pt.x - moveX) - windowRect.left, 
								windowRect.bottom + (pt.y - moveY) - windowRect.top, TRUE);

					moveX = pt.x;
					moveY = pt.y;
				}
			}
			else {

				if(maintain_aspect_ratio) {

					if(windowRect.right + (pt.x - moveX) - windowRect.left >= skin->DefaultWidth() &&
					   windowRect.bottom + (pt.y - moveY) - windowRect.top >= skin->DefaultHeight()) {
						
						MoveWindow( hwnd, windowRect.left, 
								windowRect.top, 
								windowRect.right + (pt.x - moveX) - windowRect.left, 
								windowRect.bottom + (pt.x - moveX)*playback->Height()/playback->Width() - windowRect.top, TRUE);

						moveX = pt.x;
						moveY = pt.y;
					}
				}
				else {

					if(windowRect.right + (pt.x - moveX) - windowRect.left >= skin->DefaultWidth() &&
					   windowRect.bottom + (pt.y - moveY) - windowRect.top >= skin->DefaultHeight()) {
						MoveWindow( hwnd, windowRect.left, 
									windowRect.top, 
									windowRect.right + (pt.x - moveX) - windowRect.left, 
									windowRect.bottom + (pt.y - moveY) - windowRect.top, TRUE);
					}

					moveX = pt.x;
					moveY = pt.y;
				}
			}

			GetWindowRect(hwnd, &windowRect);

			}

		}

		if(action == ACTION_VOLUME_CURSOR) {

			moveX = LOWORD(lParam); 
			skin->SetVolumeCursorX(hwnd, moveX);

			playback->SetVolume(skin->GetVolume());
		}

		if(action == ACTION_PROGRESS_CURSOR) {

			moveX = LOWORD(lParam); 
			skin->SetProgressCursorX(hwnd, moveX);
		}

		break;

	case WM_RBUTTONDOWN:

		if(!playback->fullscreen) {

			HMENU menu;
			POINT pt1;
			int   x, y;

			pt1.x = pt1.y = 0;
			ClientToScreen(hwnd, &pt1);

			x = pt1.x + LOWORD(lParam); 
			y = pt1.y + HIWORD(lParam);

			menu = GetSubMenu(popupMenu, 0);

			TrackPopupMenu( menu, 0, x, y, 0, hwnd, NULL);
		}
		break;

	case WM_SIZE:

		RECT  clientRect, windowRect;
		POINT other_pt;

		GetClientRect(hwnd, &clientRect);
		GetWindowRect(hwnd, &windowRect);

		other_pt.x = clientRect.left;
		other_pt.y = clientRect.top;

		ClientToScreen(hwnd, &other_pt);

		pt2.x = clientRect.right;
		pt2.y = clientRect.bottom;

		ClientToScreen(hwnd, &pt2);

	    SetWindowRgn(hwnd, CreateRectRgn( other_pt.x  - windowRect.left, 
			  							  other_pt.y  - windowRect.top,
										  (windowRect.right - windowRect.left) - (windowRect.right - pt2.x),
										  (windowRect.bottom - windowRect.top) - (windowRect.bottom - pt2.y)), TRUE); 
		break;

	case WM_PAINT:

		skin->UpdateSize(hwnd);
		skin->Display(hwnd, playback);
		return 0;
		break;

	case WM_DROPFILES:

		char lpFilename[256];

		if(DragQueryFile( (HDROP) wParam, 0, lpFilename, 256) > 0)
		{
			OpenFileForPlaying(hwnd, lpFilename, INPUT_TYPE_FILE);
		}
		break;

	}

  return DefWindowProc(hwnd, Message, wParam, lParam);


}
